{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [],
   "source": [
    "class edgeList:\n",
    "    def __init__(self, inList=None):\n",
    "        self.edges=[]\n",
    "        self.vertices=[]\n",
    "\n",
    "        if inList: # is not empty\n",
    "            # Suddenly a new data type appears!\n",
    "            # Its a set!\n",
    "            # Set is an unordered data structure that doesnt allow duplicates\n",
    "            vset = set()\n",
    "            for e in inList: # loop through all strings\n",
    "                # string formatting (that we all know and love right?)\n",
    "                v1, v2 = e.split(\" \")\n",
    "                self.edges.append( (v1, v2) )\n",
    "                vset.add(v1) # When I add v1 (if v1 IS in set, nothing happens)\n",
    "                vset.add(v2)\n",
    "            self.vertices = list(vset) # LIST IS NOT SET!\n",
    "\n",
    "    def getVertices(self):\n",
    "        return self.vertices\n",
    "\n",
    "        # IF we did not have vertex list, the code above \n",
    "        # dealing with sets and adding vertices IS THE SOLUTION\n",
    "\n",
    "        # This is a data structure tradeoff!\n",
    "        # I spend memory to store a list (and time at construction)\n",
    "        # to do less work when I need to get vertex list\n",
    "    \n",
    "    def getEdges(self, v):\n",
    "        edges = []\n",
    "        # Loop through all edges\n",
    "        for e in self.edges:\n",
    "            # I want to look for vertex\n",
    "            if e[0]==v: # Then my vertex of interest is the starting node in edge\n",
    "                edges.append(e)\n",
    "            elif e[1]==v:\n",
    "                edges.append(e)\n",
    "        return edges\n",
    "    \n",
    "    def areAdjacent(self, v1, v2):\n",
    "        # Note: This is very inefficient!\n",
    "        v1_nodes = self.getEdges(v1)\n",
    "        v2_nodes = self.getEdges(v2)\n",
    "\n",
    "        # O(n^2) --> O(m^2)\n",
    "        for i1 in v1_nodes:\n",
    "            for i2 in v2_nodes:\n",
    "                if i1 == i2:\n",
    "                    return True\n",
    "        return False\n",
    "\n",
    "        # Could also have done this using a similar loop to above:\n",
    "        # Ex:\n",
    "        for e in self.edges:\n",
    "            if e[0]==v1 and e[1]==v2:\n",
    "                return True\n",
    "            elif e[0]==v2 and e[1]==v1:\n",
    "                return True\n",
    "        return False\n",
    "        \n",
    "    def insertVertex(self):\n",
    "        pass\n",
    "\n",
    "    def insertEdge(self, v1, v2):\n",
    "        pass"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[('u', 'v'), ('u', 'w'), ('v', 'w'), ('w', 'z')]\n",
      "['z', 'u', 'w', 'v']\n",
      "['z', 'u', 'w', 'v']\n",
      "[('u', 'w'), ('v', 'w'), ('w', 'z')]\n",
      "False\n",
      "\n"
     ]
    }
   ],
   "source": [
    "'''\n",
    "u v\n",
    "u w\n",
    "v w\n",
    "w z\n",
    "'''\n",
    "edgeList()\n",
    "inList = [\"u v\", \"u w\", \"v w\", \"w z\"]\n",
    "\n",
    "\n",
    "myg = edgeList(inList)\n",
    "print(myg.edges)\n",
    "print(myg.vertices)\n",
    "print(myg.getVertices())\n",
    "\n",
    "print(myg.getEdges(\"w\"))\n",
    "\n",
    "print(myg.areAdjacent(\"z\", \"v\"))\n",
    "\n",
    "print()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [],
   "source": [
    "class adjMatrix:\n",
    "    def __init__(self, inList=[]): # an input list of edges [ (A, B), (B, C), ...]\n",
    "        self.vertices={} #dictionary (key = label, value = index)\n",
    "        self.edges=[] # matrix ( N x N matrix of bools! )\n",
    "\n",
    "        # Build a set of unique vertices\n",
    "        vertSet = set()\n",
    "        for edge in inList:\n",
    "            vertSet.add(edge[0])\n",
    "            vertSet.add(edge[1])\n",
    "\n",
    "        # Define an order for our matrix!\n",
    "        # Make our vertex dictionary\n",
    "        count = 0\n",
    "        for v in vertSet:\n",
    "            print(\"{}: {}\".format(v, count))\n",
    "            # Dictionary usage: d[key] = v ADDS TO DICTIONARY\n",
    "            self.vertices[v]=count # Add it to my dictionary\n",
    "            count+=1\n",
    "\n",
    "        # Make our edge matrix (empty)\n",
    "        # This makes a N x N empty matrix\n",
    "        for i in range(len(self.vertices.keys())): # N iterations (for n vertices)\n",
    "            self.edges.append([0] * len(self.vertices.keys()))\n",
    "\n",
    "        # Fill in the edge matrix\n",
    "        for edge in inList:\n",
    "            v1 = edge[0]\n",
    "            v2 = edge[1]\n",
    "\n",
    "            i1 = self.vertices[v1]\n",
    "            i2 = self.vertices[v2]\n",
    "\n",
    "            self.edges[i1][i2]=1 # Gives us edge u v \n",
    "            # My graph is undirected (all edges are bidirectional)\n",
    "            # The presence of edge [u,v] REQUIRES adding edge [v,u]\n",
    "            self.edges[i2][i1]=1 # Gives us edge v u  \n",
    "\n",
    "    # Return a list of all vertices in my adjmatrix\n",
    "    def getVertices(self):\n",
    "        return self.vertices.keys()\n",
    "\n",
    "    # Return a list of lists for every edge that includes v\n",
    "    def getEdges(self, v):\n",
    "        out = []\n",
    "        # Look up the index position of v\n",
    "        i = self.vertices[v] # This is the index position we need to look up\n",
    "        print(i)\n",
    "        # I probably need to do something with the matrix...\n",
    "        # Yes I need to get the row of edges\n",
    "        row = self.edges[i]\n",
    "        print(row)\n",
    "        # Now I need to know which index corresponds to vertex\n",
    "        # Look up every single vertex's index position\n",
    "        for v2 in self.vertices.keys(): # This will loop through all keys!\n",
    "            i2 = self.vertices[v2]\n",
    "            if row[i2] == 1:\n",
    "                # I have an edge!\n",
    "                myEdge = [ v , v2]\n",
    "\n",
    "                out.append(myEdge)\n",
    "\n",
    "        return out\n",
    "    \n",
    "    def areAdjacent(self, v1, v2):\n",
    "        pass"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "u: 0\n",
      "w: 1\n",
      "v: 2\n",
      "z: 3\n",
      "**\n",
      "{'u': 0, 'w': 1, 'v': 2, 'z': 3}\n",
      "**\n",
      "[[0, 1, 1, 0], [1, 0, 1, 1], [1, 1, 0, 0], [0, 1, 0, 0]]\n",
      "**\n",
      "dict_keys(['u', 'w', 'v', 'z'])\n",
      "**\n",
      "1\n",
      "[1, 0, 1, 1]\n",
      "[['w', 'u'], ['w', 'v'], ['w', 'z']]\n"
     ]
    }
   ],
   "source": [
    "inList = [(\"u\", \"v\"),(\"u\", \"w\"),(\"v\", \"w\"),(\"w\", \"z\")]\n",
    "myg = adjMatrix(inList)\n",
    "print(\"**\")\n",
    "print(myg.vertices) # Expected: {'u': 0, 'v': 1, 'w': 2, 'z': 3}\n",
    "print(\"**\")\n",
    "print(myg.edges) # Expected: [[0, 1, 1, 0], [1, 0, 1, 0], [1, 1, 0, 1], [0, 0, 1, 0]]\n",
    "print(\"**\")\n",
    "print(myg.getVertices())\n",
    "print(\"**\")\n",
    "print(myg.getEdges('w')) # Expected: [ ['u', 'w'], ['w', 'z'], ['w', 'v'] ] \n",
    "#print(myg.areAdjacent('u','z'))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class adjList:\n",
    "    def __init__(self, inList=[]):\n",
    "        # The vertices are the keys in the edge map\n",
    "        self.edges={}\n",
    "\n",
    "        for edge in inList:\n",
    "            v1 = edge[0]\n",
    "            v2 = edge[1]\n",
    "            #print(edge)\n",
    "            if v1 in self.edges: # If my vertex already exists\n",
    "                self.edges[v1].append(v2)\n",
    "            else: \n",
    "                self.edges[v1] = [v2] # If my vertex doesnt exist (this edge is the FIRST I saw)\n",
    "            \n",
    "            if v2 in self.edges:\n",
    "                self.edges[v2].append(v1)\n",
    "            else:\n",
    "                self.edges[v2] = [v1]\n",
    "            \n",
    "            \n",
    "\n",
    "    # Output is a list of all vertices\n",
    "    def getVertices(self): # Input is an adjacency list\n",
    "        pass\n",
    "\n",
    "    # Output is a list of tuples containing edges\n",
    "    def getEdges(self, v): # Input is a vertex key (label)\n",
    "        pass\n",
    "\n",
    "    # A boolean -- if the edge v1, v2 exists\n",
    "    def areAdjacent(self, v1, v2): # Two input vertices (as key labels)\n",
    "        pass"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "inList = [(\"u\", \"v\"),(\"u\", \"w\"),(\"v\", \"w\"),(\"w\", \"z\")]\n",
    "myg = adjList(inList)\n",
    "print(myg.getVertices())\n",
    "print(myg.getEdges('v'))# [(u, v), (w, v)] (this could be flipped -- v, u or v, w)\n",
    "print(myg.areAdjacent(\"v\", \"u\"))\n",
    "print(myg.edges) # Expected: {'u': ['v', 'w'], 'v': ['u', 'w'], 'w': ['u', 'v', 'z'], 'z': ['w']}"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
